﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  05/27/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Data
{
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Diagnostics;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Text;
    using System.Xml;
    using System.Xml.Linq;
    using System.Xml.Resolvers;


    /// <summary>
    /// Defines various acceptable country naming styles.
    /// </summary>
    public enum CountryNameStyle
    {
        /// <summary>
        /// Normal country name.
        /// </summary>
        Normal = 0,
        /// <summary>
        /// Alternative shorter country name.
        /// </summary>
        Short = 1,
        /// <summary>
        /// Alternative country name.
        /// </summary>
        Variant = 2
    }

    /// <summary>
    /// Defines a country name definition.
    /// </summary>
    [DebuggerDisplay("Code = {Code}, Name = {Name}, Style = {Style}")]
    public class CountryCodeDefinition
    {
        /// <summary>
        /// Gets the country code.
        /// </summary>
        public string Code { get; private set; }

        /// <summary>
        /// Gets the country name.
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// Gets the country name style.
        /// </summary>
        public CountryNameStyle Style { get; private set; }

        /// <summary>
        /// Initializes the CountryCodeDefinition with a code, name and style.
        /// </summary>
        /// <param name="code">The country code.</param>
        /// <param name="name">The country name.</param>
        /// <param name="style">The country name style.</param>
        public CountryCodeDefinition(string code, string name, CountryNameStyle style)
        {
            Code = code;
            Name = name;
            Style = style;
        }
    }

    /// <summary>
    /// The CommonLocaleDataRepositoryArchive class loads locale data from an CLDR archive.
    /// </summary>
    public class CommonLocaleDataRepositoryArchive : IDisposable
    {
        #region Classes
        /// <summary>
        /// Settings defines properties which determine how the CommonLocaleDataRepositoryArchive will interpret data files.
        /// </summary>
        private class Settings
        {
            /// <summary>
            /// Gets or sets a collection of distinguishing attribute names.
            /// </summary>
            public HashSet<string> DistinguishingAttributes { get; set; }
            /// <summary>
            /// Gets or sets a collection of distinguishing element names with attribute names.
            /// </summary>
            public Dictionary<string, HashSet<string>> DistinguishingElementAttributes { get; set; }
            /// <summary>
            /// Gets or sets a collection of non-distinguishing element names with attribute names.
            /// </summary>
            public Dictionary<string, HashSet<string>> ExcludeDistinguishingElementAttributes { get; set; }

            /// <summary>
            /// 
            /// </summary>
            public Dictionary<string, string> LanguageAliases { get; set; }

            /// <summary>
            /// 
            /// </summary>
            public Dictionary<string, string> TerritoryAliases { get; set; }

            /// <summary>
            /// 
            /// </summary>
            public Dictionary<string, string> ScriptAliases { get; set; }

            /// <summary>
            /// Gets or sets likely sub-tags for locale identifiers.
            /// </summary>
            //public Dictionary<string, string> LikelySubtags { get; set; }
        }

        /// <summary>
        /// The NodeDefinition class is used for comparison while determining if elements are considered overlapping.
        /// </summary>
        private class NodeDefinition
        {
            /// <summary>
            /// Gets or sets the underlying element.
            /// </summary>
            public XElement Element
            {
                get;
                set;
            }

            /// <summary>
            /// Get or set if the element is a preferred element.
            /// </summary>
            public bool Preferred
            {
                get;
                set;
            }
        }

        /// <summary>
        /// The NodeDefinitionComparer compares NodeDefinition instances to determine if an element is a descendent.
        /// </summary>
        private class NodeDefinitionComparer : IEqualityComparer<NodeDefinition>
        {
            /// <summary>
            /// Defines the settings to use for comparison.
            /// </summary>
            private Settings _Settings;

            /// <summary>
            /// Initialize the comparer with distinguishing attributes to use for comparison.
            /// </summary>
            /// <param name="settings">Settings object with distinguishing attributes.</param>
            public NodeDefinitionComparer(Settings settings)
            {
                if (settings == null)
                    throw new ArgumentNullException("settings");

                _Settings = settings;
            }

            /// <summary>
            /// Determines if an element is a descendent.
            /// </summary>
            /// <param name="x">The first element.</param>
            /// <param name="y">The second element.</param>
            /// <returns>true if the elements are equal (one is a descendent) or have matching distinguishing attribute values; otherwise, false.</returns>
            public bool Equals(NodeDefinition x, NodeDefinition y)
            {
                XElement x_element = x.Element;
                XElement y_element = y.Element;

                string xName = x_element.Name.LocalName;
                string yName = y_element.Name.LocalName;

                //if element names match and they are not from the same document
                if (xName == yName && x.Preferred != y.Preferred)
                {
                    bool xHasAttributes = x_element.HasAttributes;
                    bool yHasAttributes = y_element.HasAttributes;

                    if (xHasAttributes && yHasAttributes)
                    {
                        //get any element exclusion or inclusions
                        HashSet<string> elementExclusions = _Settings.ExcludeDistinguishingElementAttributes.ContainsKey(xName) ? _Settings.ExcludeDistinguishingElementAttributes[xName] : null;
                        HashSet<string> elementInclusions = _Settings.DistinguishingElementAttributes.ContainsKey(xName) ? _Settings.DistinguishingElementAttributes[xName] : null;

                        //enumerate all attributes
                        foreach (XAttribute attribute in x_element.Attributes())
                        {
                            string attribute_name = attribute.Name.LocalName;

                            //check if the attribute is a distinguishing attribute:
                            if ((_Settings.DistinguishingAttributes.Contains(attribute_name) ||
                               (elementInclusions != null && elementInclusions.Contains(attribute_name))) &&
                               (elementExclusions == null || !elementExclusions.Contains(attribute_name)))
                            {
                                XAttribute y_attribute = y_element.Attribute(attribute.Name);

                                if (y_attribute != null)
                                    return attribute.Value == y_attribute.Value;
                            }
                        }
                    }
                    else if (xHasAttributes || yHasAttributes)
                        return false;

                    return true;
                }

                return false;
            }

            /// <summary>
            /// Gets a hash code for the element.
            /// </summary>
            /// <param name="obj">The node definition to generate a hash code for.</param>
            /// <returns>An System.Int32 that contains the hash code for the element name.</returns>
            public int GetHashCode(NodeDefinition obj)
            {
                return obj.Element.Name.GetHashCode();
            }
        }
        #endregion

        #region Fields
        #region Private Fields
        /// <summary>
        /// Defines settings indicating how to interpret data.
        /// </summary>
        private Settings _Settings;
        /// <summary>
        /// Gets or sets a collection of distinguishing attribute names.
        /// </summary>
        private HashSet<string> DistinguishingAttributes;
        /// <summary>
        /// Gets or sets a collection of distinguishing element names with attribute names.
        /// </summary>
        private Dictionary<string, HashSet<string>> DistinguishingElementAttributes;
        /// <summary>
        /// Gets or sets a collection of non-distinguishing element names with attribute names.
        /// </summary>
        private Dictionary<string, HashSet<string>> ExcludeDistinguishingElementAttributes;
        /// <summary>
        /// Defines the stream of archive data.
        /// </summary>
        private readonly Stream _ArchiveStream;
        /// <summary>
        /// Defines the compressed archive of locale data.
        /// </summary>
        private readonly ZipArchive _Archive;
        /// <summary>
        /// Defines the main document type definition.
        /// </summary>
        private XmlReaderSettings _MainDocumentType;
        /// <summary>
        /// Defines a cache of country codes and their respective name for each locale identifier.
        /// </summary>
        private Dictionary<string, ReadOnlyCollection<CountryCodeDefinition>> _CountryCodes;
        /// <summary>
        /// Defines a cache of locale identifiers and their respective xml documents.
        /// </summary>
        private Dictionary<string, XDocument> _Repository;
        #endregion
        #endregion

        #region Disposing
        ~CommonLocaleDataRepositoryArchive()
        {
            Disposing(false);
        }

        /// <summary>
        /// Closes and disposes the archive data stream.
        /// </summary>
        public void Dispose()
        {
            Disposing(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
        protected void Disposing(bool disposing)
        {
            if (disposing)
            {
                _Archive.Dispose();
                _ArchiveStream.Dispose();
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes from the specified archive path with settings.
        /// </summary>
        /// <param name="archivePath">The archive file to load.</param>
        /// <exception cref="FileNotFoundException">Exception raised when the archive was not found.</exception>
        /// <exception cref="InvalidFileDataException">Exception raised when expected XML data is not found in the archive.</exception>
        public CommonLocaleDataRepositoryArchive(string archivePath)
        {
            if (!File.Exists(archivePath))
                throw new FileNotFoundException("The specified archive was not found", archivePath);

            _ArchiveStream = File.Open(archivePath, FileMode.Open, FileAccess.Read, FileShare.None);
            _Archive = new ZipArchive(_ArchiveStream, ZipArchiveMode.Read);

            _Repository = new Dictionary<string, XDocument>();
            _CountryCodes = new Dictionary<string, ReadOnlyCollection<CountryCodeDefinition>>();
            _MainDocumentType = LoadDocumentTypeDefinition(_Archive, "common/dtd/ldml.dtd"); //load document type definition for data files

            LoadDistinguishingElementsAttributesAndAliases(); //load the distinguishing elements and attributes
        }
        #endregion

        #region Methods
        /// <summary>
        /// Gets a collection of country code data for the specified locale identifier.
        /// </summary>
        /// <param name="localeIdentifier">The locale to get country code data in.</param>
        /// <returns>An enumerable collection of country code data.</returns>
        public IEnumerable<CountryCodeDefinition> GetCountryCodes(string localeIdentifier)
        {
            if (localeIdentifier == null)
                throw new ArgumentNullException("localeIdentifier");

            if (localeIdentifier == string.Empty)
                throw new ArgumentException("localeIdentifier must be a valid locale identifier.", "localeIdentifier");

            if (localeIdentifier == "root")
                throw new ArgumentException("Cannot load from the root locale.", "localeIdentifier");

            if (!_CountryCodes.ContainsKey(localeIdentifier))
            {
                XDocument repository = GetDocument(localeIdentifier);
                List<CountryCodeDefinition> countries = new List<CountryCodeDefinition>();

                //validates country codes:
                Func<string, bool> validCountryCode = (code) =>
                {
                    if (code == null || code.Length != 2) //skip other regions defined with 3 digit numeric codes
                        return false;

                    //skip invalid country codes:
                    foreach (char c in code)
                        if (!((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))) //a-z || A-Z
                            return false;

                    return !string.Equals(code, "ZZ"); //ZZ = Unknown
                };

                //enumerate each territory node
                foreach (XElement node in repository.Element("ldml").Element("localeDisplayNames").Element("territories").Descendants("territory"))
                {
                    XAttribute type = node.Attribute("type");

                    if (type != null)
                    {
                        string code = type.Value; //country code

                        if (validCountryCode(code)) //validate country code
                        {
                            XAttribute alt = node.Attribute("alt");

                            if (alt == null)
                                //if no alt attribute, just add the country
                                countries.Add(new CountryCodeDefinition(code, node.Value, CountryNameStyle.Normal));
                            else
                            {
                                switch (alt.Value) //check alt attribute
                                {
                                    case "short": //short country name
                                        countries.Add(new CountryCodeDefinition(code, node.Value, CountryNameStyle.Short));
                                        break;
                                    case "variant": //variant country name
                                        countries.Add(new CountryCodeDefinition(code, node.Value, CountryNameStyle.Variant));
                                        break;
                                }
                            }
                        }
                    }
                }

                return _CountryCodes[localeIdentifier] = new ReadOnlyCollection<CountryCodeDefinition>(countries);
            }

            return _CountryCodes[localeIdentifier];
        }

        /// <summary>
        /// Gets or creates a document with the specified locale identifier.
        /// </summary>
        /// <param name="localeIdentifier">The xml file locale identifier to retrieve.</param>
        /// <returns>A document loaded or retrieved from the cache with the specified locale identifier.</returns>
        private XDocument GetDocument(string localeIdentifier)
        {
            if (_Repository.ContainsKey(localeIdentifier))
                return _Repository[localeIdentifier];
            else
                return _Repository[localeIdentifier] = LoadDocument(localeIdentifier);
        }

        #region Private XML Loading Methods
        /// <summary>
        /// Loads and returns an XDocument populated from the CLDR archive.
        /// </summary>
        /// <param name="localeIdentifier">The xml file locale identifier to load.</param>
        /// <returns>An XDocument read from the CLDR archive.</returns>
        private XDocument LoadDocument(string localeIdentifier)
        {
            NodeDefinitionComparer comparer = new NodeDefinitionComparer(_Settings);

            XDocument document;
            XElement union_element = null;

            //creates an xml document by overwriting inherited elements
            do
            {
                //get zip file entry
                ZipArchiveEntry entry = _Archive.GetEntry("common/main/" + localeIdentifier + ".xml");

                if (entry == null)
                    throw new InvalidDataException(string.Format("The data file does not contain the locale: '{0}'.", localeIdentifier));

                //load the document
                document = LoadDocument(_Archive, _MainDocumentType, entry);
                //coming elements
                union_element = union_element == null ? document.Root : UnionElements(union_element, document.Root, comparer);
                //get next locale identifier
                localeIdentifier = GetParentDocumentIdentifier(document, localeIdentifier);
            }
            while (localeIdentifier != null);

            //create a new document with the original document type and combined elements
            return new XDocument(new XDeclaration(document.Declaration), document.DocumentType, union_element);
        }

        /// <summary>
        /// Gets the parent document locale identifier.
        /// </summary>
        /// <param name="document">The document to determine the new identifier from.</param>
        /// <param name="localeIdentifier">The current locale identifier.</param>
        /// <returns>The parent locale identifier or null if root.</returns>
        private string GetParentDocumentIdentifier(XDocument document, string localeIdentifier)
        {
            if (localeIdentifier == "root")
            {
                return null;
            }
            else
            {
                XElement element = document.Element("ldml").Element("identity"), territory_element = element.Element("territory"), language_element = element.Element("language"), script_element = element.Element("script");
                string territory = territory_element == null ? null : territory_element.Attribute("type").Value, language = language_element == null ? null : language_element.Attribute("type").Value, script = script_element == null ? null : script_element.Attribute("type").Value;

                if (script == null)
                {
                    if (territory == null)
                        return "root";
                    else
                        return language;
                }
                else
                {
                    if (territory == null)
                        return language;
                    else
                        return language + "_" + script;
                }
            }
        }

        /// <summary>
        /// Loads a document from a archive.
        /// </summary>
        /// <param name="archive">The archive to load from.</param>
        /// <param name="settings">The settings to create a reader from.</param>
        /// <param name="file">The file to load.</param>
        /// <returns>An document loaded from the archive.</returns>
        private XDocument LoadDocument(ZipArchive archive, XmlReaderSettings settings, string file)
        {
            ZipArchiveEntry entry = archive.GetEntry(file); //get the data file
            return LoadDocument(archive, settings, entry);
        }

        /// <summary>
        /// Loads a document from a archive.
        /// </summary>
        /// <param name="archive">The archive to load from.</param>
        /// <param name="settings">The settings to create a reader from.</param>
        /// <param name="entry">The entry to load.</param>
        /// <returns>An document loaded from the archive.</returns>
        private XDocument LoadDocument(ZipArchive archive, XmlReaderSettings settings, ZipArchiveEntry entry)
        {
            if (entry == null)
                throw new InvalidFileDataException("The zip file does not contain expected data.");

            using (Stream mainStream = entry.Open())
            using (StreamReader mainReader = new StreamReader(mainStream, Encoding.UTF8))
            using (XmlReader xmlReader = XmlReader.Create(mainReader, settings))
                //create the reader and load the xml file
                return XDocument.Load(xmlReader);
        }

        /// <summary>
        /// Loads the specified file from an archive as a document type definition.
        /// </summary>
        /// <param name="archive">The archive to load from.</param>
        /// <param name="file">The file to load.</param>
        /// <returns>Reader settings suitable to initialize an XmlReader.</returns>
        private XmlReaderSettings LoadDocumentTypeDefinition(ZipArchive archive, string file)
        {
            ZipArchiveEntry entry = archive.GetEntry(file); //get the document type definition

            if (entry == null)
                throw new InvalidFileDataException("The zip file does not contain the DTD.");

            using (Stream docTypeStream = entry.Open()) //open the entry
            {
                //XmlPreloadedResolver to populate the document type definition
                XmlPreloadedResolver resolver = new XmlPreloadedResolver();
                resolver.Add(resolver.ResolveUri(null, "../../" + file), docTypeStream);

                return new XmlReaderSettings
                {
                    DtdProcessing = DtdProcessing.Parse,
                    ValidationType = ValidationType.DTD,
                    XmlResolver = resolver
                };
            }
        }

        /// <summary>
        /// Loads all distinguishing elements, element attributes and alias replacements from the supplemental document.
        /// </summary>
        private void LoadDistinguishingElementsAttributesAndAliases()
        {
            _Settings = new Settings
            {
                DistinguishingElementAttributes = new Dictionary<string, HashSet<string>>(),
                ExcludeDistinguishingElementAttributes = new Dictionary<string, HashSet<string>>(),
                LanguageAliases = new Dictionary<string, string>(),
                TerritoryAliases = new Dictionary<string, string>(),
                ScriptAliases = new Dictionary<string, string>()//,
                //LikelySubtags = new Dictionary<string, string>()
            };

            //load the supplemental document type
            XmlReaderSettings supplementalDTD = LoadDocumentTypeDefinition(_Archive, "common/dtd/ldmlSupplemental.dtd");

            //supplemental document defining distinguishing element attributes
            XDocument supplementalDoc = LoadDocument(_Archive, supplementalDTD, "common/supplemental/supplementalMetadata.xml");

            //likely subtags document
            //XDocument likelySubtagsDoc = LoadDocument(_Archive, supplementalDTD, "common/supplemental/likelySubtags.xml");

            //get distinguishing element
            XElement metadata = supplementalDoc.Element("supplementalData").Element("metadata"), distinguishing = metadata.Element("distinguishing");

            //get alias element
            XElement aliases = metadata.Element("alias");

            //get likely subtags
            //XElement likelySubtags = likelySubtagsDoc.Element("supplementalData").Element("likelySubtags");

            //splits attribute value:
            Func<XAttribute, string[]> splitAttribute = (attribute) =>
            {
                return attribute.Value.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            };

            //adds element attributes
            Action<Dictionary<string, HashSet<string>>, XAttribute, XAttribute> addElementAttributes = (collection, elements_attribute, attributes_attribute) =>
            {
                foreach (var element_name in splitAttribute(elements_attribute)) //get all element names
                {
                    HashSet<string> attributes;

                    //get or create the collection to add distinguishing attributes to
                    if (_Settings.ExcludeDistinguishingElementAttributes.ContainsKey(element_name))
                        attributes = collection[element_name];
                    else
                        collection.Add(element_name, attributes = new HashSet<string>());

                    //add each attribute to the element
                    foreach (var attribute_name in splitAttribute(attributes_attribute))
                        if (!attributes.Contains(attribute_name))
                            attributes.Add(attribute_name);
                }
            };

            //get distinguishing element nodes
            foreach (var node in distinguishing.Nodes())
            {
                XElement element = node as XElement;

                if (element != null)
                {
                    switch (element.Name.LocalName)
                    {
                        case "distinguishingItems":
                            XAttribute exclude_attribute = element.Attribute("exclude"), attributes_attribute = element.Attribute("attributes"), elements_attribute = element.Attribute("elements");

                            if (exclude_attribute != null && exclude_attribute.Value.Equals("true"))
                            {
                                if (elements_attribute != null)
                                    //add element specific excluding element attributes
                                    addElementAttributes(_Settings.ExcludeDistinguishingElementAttributes, elements_attribute, attributes_attribute);
                            }
                            else if (elements_attribute == null)
                            {
                                if (attributes_attribute != null)
                                    //add distinguishing attributes
                                    _Settings.DistinguishingAttributes = new HashSet<string>(splitAttribute(attributes_attribute));
                            }
                            else
                                //add element specific distinguishing element attributes
                                addElementAttributes(_Settings.DistinguishingElementAttributes, elements_attribute, attributes_attribute);

                            break;
                    }
                }
            }

            if (_Settings.DistinguishingAttributes == null)
                _Settings.DistinguishingAttributes = new HashSet<string>();

            //validates the alias and adds to the collection:
            Action<Dictionary<string, string>, XElement> addAliasElement = (collection, element) =>
            {
                XAttribute type = element.Attribute("type"), replacement = element.Attribute("replacement");

                if (type != null && replacement != null)
                    if (!collection.ContainsKey(type.Value))
                        collection.Add(type.Value, replacement.Value);
            };

            //enumerate nodes
            foreach (XNode node in aliases.Nodes())
            {
                XElement element = node as XElement;

                if (element != null)
                {
                    //add node to respective alias
                    switch (element.Name.LocalName)
                    {
                        case "languageAlias":
                            addAliasElement(_Settings.LanguageAliases, element);
                            break;
                        case "territoryAlias":
                            addAliasElement(_Settings.TerritoryAliases, element);
                            break;
                        case "scriptAlias":
                            addAliasElement(_Settings.ScriptAliases, element);
                            break;
                    }
                }
            }

            //loads likely subtags
            //foreach (XNode node in likelySubtags.Nodes())
            //{
            //    XElement element = node as XElement;

            //    if (element != null)
            //    {
            //        switch (element.Name.LocalName)
            //        {
            //            case "likelySubtag":
            //                XAttribute from = element.Attribute("from"), to = element.Attribute("to");

            //                if (from != null && to != null)
            //                    if (!_Settings.LikelySubtags.ContainsKey(from.Value))
            //                        _Settings.LikelySubtags.Add(from.Value, to.Value);

            //                break;
            //        }
            //    }
            //}
        }

        /// <summary>
        /// Combines element nodes with the specified comparer.
        /// </summary>
        /// <param name="primaryElement">The primary element.</param>
        /// <param name="secondaryElement">The secondary element.</param>
        /// <param name="comparer">The comparer to use to compare elements.</param>
        /// <returns>A new element with combined sub-nodes.</returns>
        private XElement UnionElements(XElement primaryElement, XElement secondaryElement, NodeDefinitionComparer comparer)
        {
            //Initialize the new element with the primary element name and attributes
            XElement unionedElement = new XElement(primaryElement.Name);
            foreach (XAttribute attribute in primaryElement.Attributes())
                unionedElement.SetAttributeValue(attribute.Name, attribute.Value);

            //if primary element has nodes, compare child nodes to secondary element nodes
            if (primaryElement.HasElements)
            {
                //we use a dictionary with a comparer designed to distinguish elements
                Dictionary<NodeDefinition, XElement> elements = new Dictionary<NodeDefinition, XElement>(comparer);

                //enumerate secondary nodes
                foreach (XNode node in secondaryElement.Nodes())
                {
                    XElement element = node as XElement;

                    if (element != null)
                        //add NodeDefinition as a secondary node
                        elements.Add(new NodeDefinition { Element = element, Preferred = false }, element);
                }

                //enumerate primary nodes
                foreach (XNode node in primaryElement.Nodes())
                {
                    XElement element = node as XElement;

                    if (element != null)
                    {
                        //create NodeDefinition as a primary node
                        var v = new NodeDefinition { Element = element, Preferred = true };

                        //if a compatible NodeDefinition exists
                        if (elements.ContainsKey(v))
                            //combine the elements
                            elements[v] = UnionElements(element, elements[v], comparer);
                        else
                            elements.Add(v, element);
                    }
                }

                //add elements
                foreach (XElement element in elements.Values)
                    unionedElement.Add(element);
            }
            //if the secondary element has nodes, add them
            else if (secondaryElement.HasElements)
            {
                foreach (XNode node in secondaryElement.Nodes())
                {
                    XElement element = node as XElement;

                    if (element != null)
                        unionedElement.Add(element);
                }
            }
            else if (primaryElement.Value != string.Empty) //if the primary element has text, set the new element
                unionedElement.Value = primaryElement.Value;

            return unionedElement;
        }
        #endregion
        #endregion

        //TODO: Take advantage of additional CLDR data.
    }
}
